import json
from collections import OrderedDict

import pytest

from great_expectations.core import (
    ExpectationConfiguration,
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
    expectationSuiteSchema,
)
from great_expectations.data_context.util import file_relative_path
from great_expectations.render.renderer import (
    ExpectationSuiteColumnSectionRenderer,
    ProfilingResultsColumnSectionRenderer,
    ProfilingResultsOverviewSectionRenderer,
    ValidationResultsColumnSectionRenderer,
)
from great_expectations.render.renderer.content_block import (
    ValidationResultsTableContentBlockRenderer,
)


@pytest.fixture(scope="module")
def titanic_expectations():
    with open(
        file_relative_path(__file__, "../test_sets/titanic_expectations.json")
    ) as infile:
        return expectationSuiteSchema.load(
            json.load(infile, object_pairs_hook=OrderedDict)
        )


@pytest.mark.smoketest
@pytest.mark.rendered_output
def test_render_profiling_results_column_section_renderer(titanic_validation_results):
    # Group EVRs by column
    evrs = {}
    for evr in titanic_validation_results.results:
        try:
            column = evr.expectation_config.kwargs["column"]
            if column not in evrs:
                evrs[column] = []
            evrs[column].append(evr)
        except KeyError:
            pass

    for column in evrs.keys():
        with open(
            file_relative_path(
                __file__,
                "./output/test_render_profiling_results_column_section_renderer__"
                + column
                + ".json",
            ),
            "w",
        ) as outfile:
            json.dump(
                ProfilingResultsColumnSectionRenderer()
                .render(evrs[column])
                .to_json_dict(),
                outfile,
                indent=2,
            )


@pytest.mark.smoketest
@pytest.mark.rendered_output
def test_render_expectation_suite_column_section_renderer(titanic_expectations):
    # Group expectations by column
    exp_groups = {}
    # print(json.dumps(titanic_expectations, indent=2))
    for exp in titanic_expectations.expectations:
        try:
            column = exp.kwargs["column"]
            if column not in exp_groups:
                exp_groups[column] = []
            exp_groups[column].append(exp)
        except KeyError:
            pass

    for column in exp_groups.keys():
        with open(
            file_relative_path(
                __file__,
                "./output/test_render_expectation_suite_column_section_renderer"
                + column
                + ".json",
            ),
            "w",
        ) as outfile:
            json.dump(
                ExpectationSuiteColumnSectionRenderer()
                .render(exp_groups[column])
                .to_json_dict(),
                outfile,
                indent=2,
            )


@pytest.mark.smoketest
def test_ProfilingResultsColumnSectionRenderer_render(
    titanic_profiled_evrs_1, titanic_profiled_name_column_evrs
):
    # Smoke test for titanic names
    document = ProfilingResultsColumnSectionRenderer().render(
        titanic_profiled_name_column_evrs
    )
    print(document)
    assert document != {}

    # Smoke test for titanic Ages

    # This is a janky way to fetch expectations matching a specific name from an EVR suite.
    # TODO: It will no longer be necessary once we implement ValidationResultSuite._group_evrs_by_column
    from great_expectations.render.renderer.renderer import Renderer

    evrs_by_column = Renderer()._group_evrs_by_column(titanic_profiled_evrs_1)
    print(evrs_by_column.keys())

    age_column_evrs = evrs_by_column["Age"]
    for evr in age_column_evrs:
        print(evr)

    document = ProfilingResultsColumnSectionRenderer().render(age_column_evrs)
    print(document)


def test_ProfilingResultsColumnSectionRenderer_render_header(
    titanic_profiled_name_column_evrs,
):
    content_block = (
        ProfilingResultsColumnSectionRenderer()
        ._render_header(evrs=titanic_profiled_name_column_evrs, column_type=None)
        .to_json_dict()
    )

    assert content_block["content_block_type"] == "header"
    print(content_block["header"])
    assert content_block["header"] == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "Name",
            "tooltip": {"content": "expect_column_to_exist", "placement": "top"},
            "tag": "h5",
            "styling": {"classes": ["m-0", "p-0"]},
        },
    }
    print(content_block["subheader"])
    assert content_block["subheader"] == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "Type: None",
            "tooltip": {
                "content": "expect_column_values_to_be_of_type <br>expect_column_values_to_be_in_type_list"
            },
            "tag": "h6",
            "styling": {"classes": ["mt-1", "mb-0"]},
        },
    }


def test_ProfilingResultsColumnSectionRenderer_render_header_with_unescaped_dollar_sign(
    titanic_profiled_name_column_evrs,
):
    evr_with_unescaped_dollar_sign = ExpectationValidationResult(
        success=True,
        result={"observed_value": "float64"},
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_type_list",
            kwargs={
                "column": "Car Insurance Premiums ($)",
                "type_list": [
                    "DOUBLE_PRECISION",
                    "DoubleType",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT8",
                    "FloatType",
                    "NUMERIC",
                    "float",
                ],
                "result_format": "SUMMARY",
            },
            meta={"BasicDatasetProfiler": {"confidence": "very low"}},
        ),
    )

    content_block = ProfilingResultsColumnSectionRenderer._render_header(
        [evr_with_unescaped_dollar_sign], column_type=[],
    ).to_json_dict()
    print(content_block)
    assert content_block == {
        "content_block_type": "header",
        "styling": {
            "classes": ["col-12", "p-0"],
            "header": {"classes": ["alert", "alert-secondary"]},
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Car Insurance Premiums ($$)",
                "tooltip": {"content": "expect_column_to_exist", "placement": "top"},
                "tag": "h5",
                "styling": {"classes": ["m-0", "p-0"]},
            },
        },
        "subheader": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Type: []",
                "tooltip": {
                    "content": "expect_column_values_to_be_of_type <br>expect_column_values_to_be_in_type_list"
                },
                "tag": "h6",
                "styling": {"classes": ["mt-1", "mb-0"]},
            },
        },
    }


# def test_ProfilingResultsColumnSectionRenderer_render_overview_table():
#     evrs = {}
#     ProfilingResultsColumnSectionRenderer()._render_overview_table(evrs, content_blocks)

# def test_ProfilingResultsColumnSectionRenderer_render_quantile_table():
#     evrs = {}
#     ProfilingResultsColumnSectionRenderer()._render_quantile_table(evrs, content_blocks)

# def test_ProfilingResultsColumnSectionRenderer_render_stats_table():
#     evrs = {}
#     ProfilingResultsColumnSectionRenderer()._render_stats_table(evrs, content_blocks)

# def test_ProfilingResultsColumnSectionRenderer_render_histogram(titanic_profiled_evrs_1):
#     evrs = {}
#     ProfilingResultsColumnSectionRenderer()._render_histogram(evrs, content_blocks)

# def test_ProfilingResultsColumnSectionRenderer_render_values_set():
#     evrs = {}
#     ProfilingResultsColumnSectionRenderer()._render_values_set(evrs, content_blocks)


def test_ProfilingResultsColumnSectionRenderer_render_bar_chart_table(
    titanic_profiled_evrs_1,
):
    print(titanic_profiled_evrs_1.results[0])
    distinct_values_evrs = [
        evr
        for evr in titanic_profiled_evrs_1.results
        if evr.expectation_config.expectation_type
        == "expect_column_distinct_values_to_be_in_set"
    ]

    assert len(distinct_values_evrs) == 4

    content_blocks = []
    for evr in distinct_values_evrs:
        content_blocks.append(
            ProfilingResultsColumnSectionRenderer()
            ._render_bar_chart_table(distinct_values_evrs)
            .to_json_dict()
        )

    assert len(content_blocks) == 4

    for content_block in content_blocks:
        assert content_block["content_block_type"] == "graph"
        assert set(content_block.keys()) == {
            "header",
            "content_block_type",
            "graph",
            "styling",
        }
        assert json.loads(content_block["graph"])


def test_ExpectationSuiteColumnSectionRenderer_render_header(
    titanic_profiled_name_column_expectations,
):
    (
        remaining_expectations,
        content_blocks,
    ) = ExpectationSuiteColumnSectionRenderer._render_header(
        titanic_profiled_name_column_expectations,
    )

    expected = {
        "content_block_type": "header",
        "styling": {
            "classes": ["col-12"],
            "header": {"classes": ["alert", "alert-secondary"]},
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Name",
                "tag": "h5",
                "styling": {"classes": ["m-0"]},
            },
        },
    }

    print(content_blocks.to_json_dict())

    assert content_blocks.to_json_dict() == expected

    expectation_with_unescaped_dollar_sign = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_type_list",
        kwargs={
            "column": "Car Insurance Premiums ($)",
            "type_list": [
                "DOUBLE_PRECISION",
                "DoubleType",
                "FLOAT",
                "FLOAT4",
                "FLOAT8",
                "FloatType",
                "NUMERIC",
                "float",
            ],
            "result_format": "SUMMARY",
        },
        meta={"BasicDatasetProfiler": {"confidence": "very low"}},
    )
    (
        remaining_expectations,
        content_blocks,
    ) = ExpectationSuiteColumnSectionRenderer._render_header(
        [expectation_with_unescaped_dollar_sign],
    )

    print(content_blocks.to_json_dict())
    expected = {
        "content_block_type": "header",
        "styling": {
            "classes": ["col-12"],
            "header": {"classes": ["alert", "alert-secondary"]},
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Car Insurance Premiums ($$)",
                "tag": "h5",
                "styling": {"classes": ["m-0"]},
            },
        },
    }
    assert content_blocks.to_json_dict() == expected


def test_ExpectationSuiteColumnSectionRenderer_expectation_with_markdown_meta_notes():
    expectation_with_markdown_meta_notes = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_type_list",
        kwargs={
            "column": "Car Insurance Premiums ($)",
            "type_list": [
                "DOUBLE_PRECISION",
                "DoubleType",
                "FLOAT",
                "FLOAT4",
                "FLOAT8",
                "FloatType",
                "NUMERIC",
                "float",
            ],
            "result_format": "SUMMARY",
        },
        meta={
            "BasicDatasetProfiler": {"confidence": "very low"},
            "notes": {
                "format": "markdown",
                "content": [
                    "#### These are expectation notes \n - you can use markdown \n - or just strings"
                ],
            },
        },
    )
    expectations = [expectation_with_markdown_meta_notes]
    expected_result_json = {
        "content_blocks": [
            {
                "content_block_type": "header",
                "styling": {
                    "classes": ["col-12"],
                    "header": {"classes": ["alert", "alert-secondary"]},
                },
                "header": {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Car Insurance Premiums ($$)",
                        "tag": "h5",
                        "styling": {"classes": ["m-0"]},
                    },
                },
            },
            {
                "content_block_type": "bullet_list",
                "styling": {"classes": ["col-12"]},
                "bullet_list": [
                    [
                        {
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "value types must belong to this set: $v__0"
                                " $v__1 $v__2 $v__3 $v__4 $v__5 $v__6 $v__7.",
                                "params": {
                                    "column": "Car Insurance Premiums ($)",
                                    "type_list": [
                                        "DOUBLE_PRECISION",
                                        "DoubleType",
                                        "FLOAT",
                                        "FLOAT4",
                                        "FLOAT8",
                                        "FloatType",
                                        "NUMERIC",
                                        "float",
                                    ],
                                    "result_format": "SUMMARY",
                                    "mostly": None,
                                    "row_condition": None,
                                    "condition_parser": None,
                                    "v__0": "DOUBLE_PRECISION",
                                    "v__1": "DoubleType",
                                    "v__2": "FLOAT",
                                    "v__3": "FLOAT4",
                                    "v__4": "FLOAT8",
                                    "v__5": "FloatType",
                                    "v__6": "NUMERIC",
                                    "v__7": "float",
                                },
                                "styling": {
                                    "default": {
                                        "classes": ["badge", "badge-secondary"]
                                    },
                                    "params": {
                                        "column": {
                                            "classes": ["badge", "badge-primary"]
                                        }
                                    },
                                },
                            },
                        },
                        {
                            "content_block_type": "collapse",
                            "styling": {
                                "body": {"classes": ["card", "card-body", "p-1"]},
                                "parent": {"styles": {"list-style-type": "none"}},
                            },
                            "collapse_toggle_link": {
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": "$icon",
                                    "params": {"icon": ""},
                                    "styling": {
                                        "params": {
                                            "icon": {
                                                "classes": [
                                                    "fas",
                                                    "fa-comment",
                                                    "text-info",
                                                ],
                                                "tag": "i",
                                            }
                                        }
                                    },
                                },
                            },
                            "collapse": [
                                {
                                    "content_block_type": "text",
                                    "styling": {
                                        "classes": ["col-12", "mt-2", "mb-2"],
                                        "parent": {
                                            "styles": {"list-style-type": "none"}
                                        },
                                    },
                                    "subheader": "Notes:",
                                    "text": [
                                        {
                                            "content_block_type": "markdown",
                                            "styling": {"parent": {}},
                                            "markdown": "#### These are expectation notes \n - "
                                            "you can use markdown \n - or just strings",
                                        }
                                    ],
                                }
                            ],
                            "inline_link": True,
                        },
                    ],
                    {
                        "content_block_type": "string_template",
                        "styling": {"parent": {"styles": {"list-style-type": "none"}}},
                        "string_template": {
                            "template": "",
                            "tag": "hr",
                            "styling": {"classes": ["mt-1", "mb-1"]},
                        },
                    },
                ],
            },
        ],
        "section_name": "Car Insurance Premiums ($)",
    }

    result_json = (
        ExpectationSuiteColumnSectionRenderer().render(expectations).to_json_dict()
    )
    print(result_json)
    assert result_json == expected_result_json


def test_ExpectationSuiteColumnSectionRenderer_expectation_with_string_list_meta_notes_in_dict():
    expectation_with_string_notes_list_in_dict = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_type_list",
        kwargs={
            "column": "Car Insurance Premiums ($)",
            "type_list": [
                "DOUBLE_PRECISION",
                "DoubleType",
                "FLOAT",
                "FLOAT4",
                "FLOAT8",
                "FloatType",
                "NUMERIC",
                "float",
            ],
            "result_format": "SUMMARY",
        },
        meta={
            "BasicDatasetProfiler": {"confidence": "very low"},
            "notes": {
                "format": "string",
                "content": [
                    "This is a",
                    "string list,",
                    "assigned to the 'content' key of a notes dict.",
                    "Cool",
                    "huh?",
                ],
            },
        },
    )
    expectations = [expectation_with_string_notes_list_in_dict]
    expected_result_json = {
        "content_blocks": [
            {
                "content_block_type": "header",
                "styling": {
                    "classes": ["col-12"],
                    "header": {"classes": ["alert", "alert-secondary"]},
                },
                "header": {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Car Insurance Premiums ($$)",
                        "tag": "h5",
                        "styling": {"classes": ["m-0"]},
                    },
                },
            },
            {
                "content_block_type": "bullet_list",
                "styling": {"classes": ["col-12"]},
                "bullet_list": [
                    [
                        {
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "value types must belong to this set: $v__0 $v__1 "
                                "$v__2 $v__3 $v__4 $v__5 $v__6 $v__7.",
                                "params": {
                                    "column": "Car Insurance Premiums ($)",
                                    "type_list": [
                                        "DOUBLE_PRECISION",
                                        "DoubleType",
                                        "FLOAT",
                                        "FLOAT4",
                                        "FLOAT8",
                                        "FloatType",
                                        "NUMERIC",
                                        "float",
                                    ],
                                    "result_format": "SUMMARY",
                                    "mostly": None,
                                    "row_condition": None,
                                    "condition_parser": None,
                                    "v__0": "DOUBLE_PRECISION",
                                    "v__1": "DoubleType",
                                    "v__2": "FLOAT",
                                    "v__3": "FLOAT4",
                                    "v__4": "FLOAT8",
                                    "v__5": "FloatType",
                                    "v__6": "NUMERIC",
                                    "v__7": "float",
                                },
                                "styling": {
                                    "default": {
                                        "classes": ["badge", "badge-secondary"]
                                    },
                                    "params": {
                                        "column": {
                                            "classes": ["badge", "badge-primary"]
                                        }
                                    },
                                },
                            },
                        },
                        {
                            "content_block_type": "collapse",
                            "styling": {
                                "body": {"classes": ["card", "card-body", "p-1"]},
                                "parent": {"styles": {"list-style-type": "none"}},
                            },
                            "collapse_toggle_link": {
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": "$icon",
                                    "params": {"icon": ""},
                                    "styling": {
                                        "params": {
                                            "icon": {
                                                "classes": [
                                                    "fas",
                                                    "fa-comment",
                                                    "text-info",
                                                ],
                                                "tag": "i",
                                            }
                                        }
                                    },
                                },
                            },
                            "collapse": [
                                {
                                    "content_block_type": "text",
                                    "styling": {
                                        "classes": ["col-12", "mt-2", "mb-2"],
                                        "parent": {
                                            "styles": {"list-style-type": "none"}
                                        },
                                    },
                                    "subheader": "Notes:",
                                    "text": [
                                        "This is a",
                                        "string list,",
                                        "assigned to the 'content' key of a notes dict.",
                                        "Cool",
                                        "huh?",
                                    ],
                                }
                            ],
                            "inline_link": True,
                        },
                    ],
                    {
                        "content_block_type": "string_template",
                        "styling": {"parent": {"styles": {"list-style-type": "none"}}},
                        "string_template": {
                            "template": "",
                            "tag": "hr",
                            "styling": {"classes": ["mt-1", "mb-1"]},
                        },
                    },
                ],
            },
        ],
        "section_name": "Car Insurance Premiums ($)",
    }

    result_json = (
        ExpectationSuiteColumnSectionRenderer().render(expectations).to_json_dict()
    )
    print(result_json)
    assert result_json == expected_result_json


def test_ExpectationSuiteColumnSectionRenderer_expectation_with_single_string_meta_note_in_dict():
    expectation_with_single_string_note_in_dict = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_type_list",
        kwargs={
            "column": "Car Insurance Premiums ($)",
            "type_list": [
                "DOUBLE_PRECISION",
                "DoubleType",
                "FLOAT",
                "FLOAT4",
                "FLOAT8",
                "FloatType",
                "NUMERIC",
                "float",
            ],
            "result_format": "SUMMARY",
        },
        meta={
            "BasicDatasetProfiler": {"confidence": "very low"},
            "notes": {
                "format": "string",
                "content": "This is just a single string, assigned to the 'content' key of a notes dict.",
            },
        },
    )
    expectations = [expectation_with_single_string_note_in_dict]
    expected_result_json = {
        "content_blocks": [
            {
                "content_block_type": "header",
                "styling": {
                    "classes": ["col-12"],
                    "header": {"classes": ["alert", "alert-secondary"]},
                },
                "header": {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Car Insurance Premiums ($$)",
                        "tag": "h5",
                        "styling": {"classes": ["m-0"]},
                    },
                },
            },
            {
                "content_block_type": "bullet_list",
                "styling": {"classes": ["col-12"]},
                "bullet_list": [
                    [
                        {
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "value types must belong to this set: $v__0 $v__1 "
                                "$v__2 $v__3 $v__4 $v__5 $v__6 $v__7.",
                                "params": {
                                    "column": "Car Insurance Premiums ($)",
                                    "type_list": [
                                        "DOUBLE_PRECISION",
                                        "DoubleType",
                                        "FLOAT",
                                        "FLOAT4",
                                        "FLOAT8",
                                        "FloatType",
                                        "NUMERIC",
                                        "float",
                                    ],
                                    "result_format": "SUMMARY",
                                    "mostly": None,
                                    "row_condition": None,
                                    "condition_parser": None,
                                    "v__0": "DOUBLE_PRECISION",
                                    "v__1": "DoubleType",
                                    "v__2": "FLOAT",
                                    "v__3": "FLOAT4",
                                    "v__4": "FLOAT8",
                                    "v__5": "FloatType",
                                    "v__6": "NUMERIC",
                                    "v__7": "float",
                                },
                                "styling": {
                                    "default": {
                                        "classes": ["badge", "badge-secondary"]
                                    },
                                    "params": {
                                        "column": {
                                            "classes": ["badge", "badge-primary"]
                                        }
                                    },
                                },
                            },
                        },
                        {
                            "content_block_type": "collapse",
                            "styling": {
                                "body": {"classes": ["card", "card-body", "p-1"]},
                                "parent": {"styles": {"list-style-type": "none"}},
                            },
                            "collapse_toggle_link": {
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": "$icon",
                                    "params": {"icon": ""},
                                    "styling": {
                                        "params": {
                                            "icon": {
                                                "classes": [
                                                    "fas",
                                                    "fa-comment",
                                                    "text-info",
                                                ],
                                                "tag": "i",
                                            }
                                        }
                                    },
                                },
                            },
                            "collapse": [
                                {
                                    "content_block_type": "text",
                                    "styling": {
                                        "classes": ["col-12", "mt-2", "mb-2"],
                                        "parent": {
                                            "styles": {"list-style-type": "none"}
                                        },
                                    },
                                    "subheader": "Notes:",
                                    "text": [
                                        "This is just a single string, assigned to the 'content' key of a notes dict."
                                    ],
                                }
                            ],
                            "inline_link": True,
                        },
                    ],
                    {
                        "content_block_type": "string_template",
                        "styling": {"parent": {"styles": {"list-style-type": "none"}}},
                        "string_template": {
                            "template": "",
                            "tag": "hr",
                            "styling": {"classes": ["mt-1", "mb-1"]},
                        },
                    },
                ],
            },
        ],
        "section_name": "Car Insurance Premiums ($)",
    }

    result_json = (
        ExpectationSuiteColumnSectionRenderer().render(expectations).to_json_dict()
    )
    print(result_json)
    assert result_json == expected_result_json


def test_ExpectationSuiteColumnSectionRenderer_expectation_with_string_list_meta_notes():
    expectation_with_string_list_note = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_type_list",
        kwargs={
            "column": "Car Insurance Premiums ($)",
            "type_list": [
                "DOUBLE_PRECISION",
                "DoubleType",
                "FLOAT",
                "FLOAT4",
                "FLOAT8",
                "FloatType",
                "NUMERIC",
                "float",
            ],
            "result_format": "SUMMARY",
        },
        meta={
            "BasicDatasetProfiler": {"confidence": "very low"},
            "notes": ["This is a list", "of strings", "assigned to the notes", "key."],
        },
    )
    expectations = [expectation_with_string_list_note]
    expected_result_json = {
        "content_blocks": [
            {
                "content_block_type": "header",
                "styling": {
                    "classes": ["col-12"],
                    "header": {"classes": ["alert", "alert-secondary"]},
                },
                "header": {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Car Insurance Premiums ($$)",
                        "tag": "h5",
                        "styling": {"classes": ["m-0"]},
                    },
                },
            },
            {
                "content_block_type": "bullet_list",
                "styling": {"classes": ["col-12"]},
                "bullet_list": [
                    [
                        {
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "value types must belong to this set: $v__0 $v__1 "
                                "$v__2 $v__3 $v__4 $v__5 $v__6 $v__7.",
                                "params": {
                                    "column": "Car Insurance Premiums ($)",
                                    "type_list": [
                                        "DOUBLE_PRECISION",
                                        "DoubleType",
                                        "FLOAT",
                                        "FLOAT4",
                                        "FLOAT8",
                                        "FloatType",
                                        "NUMERIC",
                                        "float",
                                    ],
                                    "result_format": "SUMMARY",
                                    "mostly": None,
                                    "row_condition": None,
                                    "condition_parser": None,
                                    "v__0": "DOUBLE_PRECISION",
                                    "v__1": "DoubleType",
                                    "v__2": "FLOAT",
                                    "v__3": "FLOAT4",
                                    "v__4": "FLOAT8",
                                    "v__5": "FloatType",
                                    "v__6": "NUMERIC",
                                    "v__7": "float",
                                },
                                "styling": {
                                    "default": {
                                        "classes": ["badge", "badge-secondary"]
                                    },
                                    "params": {
                                        "column": {
                                            "classes": ["badge", "badge-primary"]
                                        }
                                    },
                                },
                            },
                        },
                        {
                            "content_block_type": "collapse",
                            "styling": {
                                "body": {"classes": ["card", "card-body", "p-1"]},
                                "parent": {"styles": {"list-style-type": "none"}},
                            },
                            "collapse_toggle_link": {
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": "$icon",
                                    "params": {"icon": ""},
                                    "styling": {
                                        "params": {
                                            "icon": {
                                                "classes": [
                                                    "fas",
                                                    "fa-comment",
                                                    "text-info",
                                                ],
                                                "tag": "i",
                                            }
                                        }
                                    },
                                },
                            },
                            "collapse": [
                                {
                                    "content_block_type": "text",
                                    "styling": {
                                        "classes": ["col-12", "mt-2", "mb-2"],
                                        "parent": {
                                            "styles": {"list-style-type": "none"}
                                        },
                                    },
                                    "subheader": "Notes:",
                                    "text": [
                                        "This is a list",
                                        "of strings",
                                        "assigned to the notes",
                                        "key.",
                                    ],
                                }
                            ],
                            "inline_link": True,
                        },
                    ],
                    {
                        "content_block_type": "string_template",
                        "styling": {"parent": {"styles": {"list-style-type": "none"}}},
                        "string_template": {
                            "template": "",
                            "tag": "hr",
                            "styling": {"classes": ["mt-1", "mb-1"]},
                        },
                    },
                ],
            },
        ],
        "section_name": "Car Insurance Premiums ($)",
    }

    result_json = (
        ExpectationSuiteColumnSectionRenderer().render(expectations).to_json_dict()
    )
    print(result_json)
    assert result_json == expected_result_json


def test_ExpectationSuiteColumnSectionRenderer_expectation_with_single_string_meta_note():
    expectation_with_single_string_note = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_type_list",
        kwargs={
            "column": "Car Insurance Premiums ($)",
            "type_list": [
                "DOUBLE_PRECISION",
                "DoubleType",
                "FLOAT",
                "FLOAT4",
                "FLOAT8",
                "FloatType",
                "NUMERIC",
                "float",
            ],
            "result_format": "SUMMARY",
        },
        meta={
            "BasicDatasetProfiler": {"confidence": "very low"},
            "notes": "This is a single string assigned to the 'notes' key.",
        },
    )
    expectations = [expectation_with_single_string_note]
    expected_result_json = {
        "content_blocks": [
            {
                "content_block_type": "header",
                "styling": {
                    "classes": ["col-12"],
                    "header": {"classes": ["alert", "alert-secondary"]},
                },
                "header": {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Car Insurance Premiums ($$)",
                        "tag": "h5",
                        "styling": {"classes": ["m-0"]},
                    },
                },
            },
            {
                "content_block_type": "bullet_list",
                "styling": {"classes": ["col-12"]},
                "bullet_list": [
                    [
                        {
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "value types must belong to this set: $v__0"
                                " $v__1 $v__2 $v__3 $v__4 $v__5 $v__6 $v__7.",
                                "params": {
                                    "column": "Car Insurance Premiums ($)",
                                    "type_list": [
                                        "DOUBLE_PRECISION",
                                        "DoubleType",
                                        "FLOAT",
                                        "FLOAT4",
                                        "FLOAT8",
                                        "FloatType",
                                        "NUMERIC",
                                        "float",
                                    ],
                                    "result_format": "SUMMARY",
                                    "mostly": None,
                                    "row_condition": None,
                                    "condition_parser": None,
                                    "v__0": "DOUBLE_PRECISION",
                                    "v__1": "DoubleType",
                                    "v__2": "FLOAT",
                                    "v__3": "FLOAT4",
                                    "v__4": "FLOAT8",
                                    "v__5": "FloatType",
                                    "v__6": "NUMERIC",
                                    "v__7": "float",
                                },
                                "styling": {
                                    "default": {
                                        "classes": ["badge", "badge-secondary"]
                                    },
                                    "params": {
                                        "column": {
                                            "classes": ["badge", "badge-primary"]
                                        }
                                    },
                                },
                            },
                        },
                        {
                            "content_block_type": "collapse",
                            "styling": {
                                "body": {"classes": ["card", "card-body", "p-1"]},
                                "parent": {"styles": {"list-style-type": "none"}},
                            },
                            "collapse_toggle_link": {
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": "$icon",
                                    "params": {"icon": ""},
                                    "styling": {
                                        "params": {
                                            "icon": {
                                                "classes": [
                                                    "fas",
                                                    "fa-comment",
                                                    "text-info",
                                                ],
                                                "tag": "i",
                                            }
                                        }
                                    },
                                },
                            },
                            "collapse": [
                                {
                                    "content_block_type": "text",
                                    "styling": {
                                        "classes": ["col-12", "mt-2", "mb-2"],
                                        "parent": {
                                            "styles": {"list-style-type": "none"}
                                        },
                                    },
                                    "subheader": "Notes:",
                                    "text": [
                                        "This is a single string assigned to the 'notes' key."
                                    ],
                                }
                            ],
                            "inline_link": True,
                        },
                    ],
                    {
                        "content_block_type": "string_template",
                        "styling": {"parent": {"styles": {"list-style-type": "none"}}},
                        "string_template": {
                            "template": "",
                            "tag": "hr",
                            "styling": {"classes": ["mt-1", "mb-1"]},
                        },
                    },
                ],
            },
        ],
        "section_name": "Car Insurance Premiums ($)",
    }

    result_json = (
        ExpectationSuiteColumnSectionRenderer().render(expectations).to_json_dict()
    )
    print(result_json)
    assert result_json == expected_result_json


def test_ExpectationSuiteColumnSectionRenderer_render_bullet_list(
    titanic_profiled_name_column_expectations,
):
    (
        remaining_expectations,
        content_block,
    ) = ExpectationSuiteColumnSectionRenderer()._render_bullet_list(
        titanic_profiled_name_column_expectations,
    )

    stringified_dump = json.dumps(content_block.to_json_dict())

    assert content_block.content_block_type == "bullet_list"
    assert len(content_block.bullet_list) == 8
    assert "value types must belong to this set" in stringified_dump
    assert "may have any number of unique values" in stringified_dump
    assert "may have any fraction of unique values" in stringified_dump
    assert (
        "values must not be null, at least $mostly_pct % of the time."
        in stringified_dump
    )


def test_ValidationResultsColumnSectionRenderer_render_header(
    titanic_profiled_name_column_evrs,
):
    (
        remaining_evrs,
        content_block,
    ) = ValidationResultsColumnSectionRenderer._render_header(
        validation_results=titanic_profiled_name_column_evrs,
    )
    print(content_block.to_json_dict())
    assert content_block.to_json_dict() == {
        "content_block_type": "header",
        "styling": {
            "classes": ["col-12", "p-0"],
            "header": {"classes": ["alert", "alert-secondary"]},
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Name",
                "tag": "h5",
                "styling": {"classes": ["m-0"]},
            },
        },
    }


def test_ValidationResultsColumnSectionRenderer_render_header_evr_with_unescaped_dollar_sign(
    titanic_profiled_name_column_evrs,
):
    evr_with_unescaped_dollar_sign = ExpectationValidationResult(
        success=True,
        result={
            "element_count": 1313,
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "unexpected_percent_nonmissing": 0.0,
            "partial_unexpected_list": [],
            "partial_unexpected_index_list": [],
            "partial_unexpected_counts": [],
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_type_list",
            kwargs={
                "column": "Name ($)",
                "type_list": ["CHAR", "StringType", "TEXT", "VARCHAR", "str", "string"],
                "result_format": "SUMMARY",
            },
        ),
    )

    (
        remaining_evrs,
        content_block,
    ) = ValidationResultsColumnSectionRenderer._render_header(
        validation_results=[evr_with_unescaped_dollar_sign],
    )

    print(content_block.to_json_dict())

    assert content_block.to_json_dict() == {
        "content_block_type": "header",
        "styling": {
            "classes": ["col-12", "p-0"],
            "header": {"classes": ["alert", "alert-secondary"]},
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Name ($$)",
                "tag": "h5",
                "styling": {"classes": ["m-0"]},
            },
        },
    }


# noinspection PyPep8Naming
def test_ValidationResultsColumnSectionRenderer_render_table(
    titanic_profiled_name_column_evrs,
):
    (
        remaining_evrs,
        content_block,
    ) = ValidationResultsColumnSectionRenderer()._render_table(
        validation_results=titanic_profiled_name_column_evrs,
    )

    content_block_stringified = json.dumps(content_block.to_json_dict())

    assert content_block.content_block_type == "table"
    assert len(content_block.table) == 6
    assert content_block_stringified.count("$icon") == 6
    assert (
        "value types must belong to this set: $v__0 $v__1 $v__2 $v__3 $v__4 $v__5 $v__6"
        in content_block_stringified
    )
    assert "may have any number of unique values." in content_block_stringified
    assert "may have any fraction of unique values." in content_block_stringified
    assert (
        "values must not be null, at least $mostly_pct % of the time."
        in content_block_stringified
    )
    assert "values must belong to this set: [ ]." in content_block_stringified
    assert (
        "\\n\\n$unexpected_count unexpected values found. $unexpected_percent of $element_count total rows."
        in content_block_stringified
    )
    assert (
        "values must not match this regular expression: $regex."
        in content_block_stringified
    )
    assert (
        "\\n\\n$unexpected_count unexpected values found. $unexpected_percent of $element_count total rows."
        in content_block_stringified
    )


# noinspection PyPep8Naming
def test_ValidationResultsTableContentBlockRenderer_generate_expectation_row_happy_path():
    evr = ExpectationValidationResult(
        success=True,
        result={
            "observed_value": True,
            "element_count": 162,
            "missing_count": 153,
            "missing_percent": 94.44444444444444,
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_min_to_be_between",
            kwargs={
                "column": "live",
                "min_value": None,
                "max_value": None,
                "result_format": "SUMMARY",
            },
            meta={"BasicDatasetProfiler": {"confidence": "very low"}},
        ),
    )
    result = ValidationResultsTableContentBlockRenderer.render([evr]).to_json_dict()
    print(result)

    # Note: A better approach to testing would separate out styling into a separate test.
    assert result == {
        "content_block_type": "table",
        "styling": {
            "body": {"classes": ["table"]},
            "classes": [
                "ml-2",
                "mr-2",
                "mt-0",
                "mb-0",
                "table-responsive",
                "hide-succeeded-validations-column-section-target-child",
            ],
        },
        "table": [
            [
                {
                    "content_block_type": "string_template",
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target-child"]
                        }
                    },
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                },
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$column minimum value may have any numerical value.",
                        "params": {
                            "column": "live",
                            "min_value": None,
                            "max_value": None,
                            "result_format": "SUMMARY",
                            "parse_strings_as_datetimes": None,
                            "row_condition": None,
                            "condition_parser": None,
                            "strict_max": None,
                            "strict_min": None,
                        },
                        "styling": {
                            "default": {"classes": ["badge", "badge-secondary"]},
                            "params": {
                                "column": {"classes": ["badge", "badge-primary"]}
                            },
                        },
                    },
                },
                "True",
            ]
        ],
        "header_row": ["Status", "Expectation", "Observed Value"],
        "header_row_options": {"Status": {"sortable": True}},
        "table_options": {"search": True, "icon-size": "sm"},
    }


# noinspection PyPep8Naming
def test_ProfilingResultsOverviewSectionRenderer_empty_type_list():
    # This rather specific test is a reaction to the error documented in #679
    validation = ExpectationSuiteValidationResult(
        results=[
            ExpectationValidationResult(
                success=True,
                result={
                    "observed_value": "VARIANT",  # Note this is NOT a recognized type by many backends
                },
                exception_info={
                    "raised_exception": False,
                    "exception_message": None,
                    "exception_traceback": None,
                },
                expectation_config=ExpectationConfiguration(
                    expectation_type="expect_column_values_to_be_in_type_list",
                    kwargs={
                        "column": "live",
                        "type_list": None,
                        "result_format": "SUMMARY",
                    },
                    meta={"BasicDatasetProfiler": {"confidence": "very low"}},
                ),
            )
        ]
    )

    result = ProfilingResultsOverviewSectionRenderer().render(validation)

    # Find the variable types content block:
    types_table = [
        block.table
        for block in result.content_blocks
        if block.content_block_type == "table"
        and block.header.string_template["template"] == "Variable types"
    ][0]
    assert ["unknown", "1"] in types_table



import pytest

from great_expectations.core import ExpectationConfiguration, ExpectationSuite
from great_expectations.exceptions import (
    DataContextError,
    InvalidExpectationConfigurationError,
)


@pytest.fixture
def empty_suite():
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def baseline_suite(exp1, exp2):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1, exp2],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def exp1():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp2():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp3():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp4():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp5():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3], "result_format": "COMPLETE"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp6():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp7():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3, 4]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp8():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def table_exp1():
    return ExpectationConfiguration(
        expectation_type="expect_table_columns_to_match_ordered_list",
        kwargs={"value": ["a", "b", "c"]},
    )


@pytest.fixture
def table_exp2():
    return ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_be_between",
        kwargs={"min_value": 0, "max_value": 1},
    )


@pytest.fixture
def table_exp3():
    return ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_equal", kwargs={"value": 1}
    )


@pytest.fixture
def column_pair_expectation():
    return ExpectationConfiguration(
        expectation_type="expect_column_pair_values_to_be_in_set",
        kwargs={
            "column_A": "1",
            "column_B": "b",
            "value_set": [(1, 1), (2, 2)],
            "result_format": "BASIC",
        },
    )


@pytest.fixture
def single_expectation_suite(exp1):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def different_suite(exp1, exp4):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1, exp4],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def domain_success_runtime_suite(exp1, exp2, exp3, exp4, exp5):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1, exp2, exp3, exp4, exp5],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def suite_with_table_and_column_expectations(
    exp1, exp2, exp3, exp4, column_pair_expectation, table_exp1, table_exp2, table_exp3
):
    suite = ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[
            exp1,
            exp2,
            exp3,
            exp4,
            column_pair_expectation,
            table_exp1,
            table_exp2,
            table_exp3,
        ],
        meta={"notes": "This is an expectation suite."},
    )
    assert suite.expectations == [
        exp1,
        exp2,
        exp3,
        exp4,
        column_pair_expectation,
        table_exp1,
        table_exp2,
        table_exp3,
    ]
    return suite


@pytest.fixture
def suite_with_column_pair_and_table_expectations(
    table_exp1, table_exp2, table_exp3, column_pair_expectation
):
    suite = ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[column_pair_expectation, table_exp1, table_exp2, table_exp3,],
        meta={"notes": "This is an expectation suite."},
    )
    assert suite.expectations == [
        column_pair_expectation,
        table_exp1,
        table_exp2,
        table_exp3,
    ]
    return suite


def test_find_expectation_indexes_on_empty_suite(exp1, empty_suite):
    assert empty_suite.find_expectation_indexes(exp1, "domain") == []


def test_find_expectation_indexes(
    exp1, exp2, exp3, exp4, exp5, domain_success_runtime_suite, single_expectation_suite
):
    assert domain_success_runtime_suite.find_expectation_indexes(exp4, "domain") == [
        1,
        2,
        3,
        4,
    ]
    assert domain_success_runtime_suite.find_expectation_indexes(exp4, "success") == [
        3,
        4,
    ]
    assert domain_success_runtime_suite.find_expectation_indexes(exp4, "runtime") == [3]

    assert single_expectation_suite.find_expectation_indexes(exp4, "runtime") == []

    with pytest.raises(InvalidExpectationConfigurationError):
        domain_success_runtime_suite.remove_expectation(
            "not an expectation", match_type="runtime"
        )

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(
            exp1, match_type="not a match_type"
        )


def test_find_expectations(
    exp1, exp2, exp3, exp4, exp5, domain_success_runtime_suite, single_expectation_suite
):
    expectation_to_find1 = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "COMPLETE"},
    )

    expectation_to_find2 = ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3], "result_format": "COMPLETE"},
    )

    assert domain_success_runtime_suite.find_expectations(
        expectation_to_find1, "domain"
    ) == [exp2, exp3, exp4, exp5]

    assert domain_success_runtime_suite.find_expectations(
        expectation_to_find1, "success"
    ) == [exp2, exp3]

    assert domain_success_runtime_suite.find_expectations(
        expectation_to_find2, "runtime"
    ) == [exp5]

    assert (
        domain_success_runtime_suite.find_expectations(expectation_to_find1, "runtime")
        == []
    )


def test_remove_expectation(
    exp1, exp2, exp3, exp4, exp5, single_expectation_suite, domain_success_runtime_suite
):
    domain_success_runtime_suite.remove_expectation(
        exp5, match_type="runtime", remove_multiple_matches=False
    )  # remove one matching expectation

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(exp5, match_type="runtime")
    assert domain_success_runtime_suite.find_expectation_indexes(exp4, "domain") == [
        1,
        2,
        3,
    ]
    assert domain_success_runtime_suite.find_expectation_indexes(exp4, "success") == [3]

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(
            exp4, match_type="domain", remove_multiple_matches=False
        )

    # remove 3 matching expectations
    domain_success_runtime_suite.remove_expectation(
        exp4, match_type="domain", remove_multiple_matches=True
    )

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(exp2, match_type="runtime")
    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(exp3, match_type="runtime")

    assert domain_success_runtime_suite.find_expectation_indexes(
        exp1, match_type="domain"
    ) == [0]
    assert domain_success_runtime_suite.isEquivalentTo(single_expectation_suite)


def test_patch_expectation_replace(exp5, exp6, domain_success_runtime_suite):
    assert domain_success_runtime_suite.expectations[4] is exp5

    assert not domain_success_runtime_suite.expectations[4].isEquivalentTo(
        exp6, match_type="success"
    )
    domain_success_runtime_suite.patch_expectation(
        exp5, op="replace", path="/value_set", value=[1, 2], match_type="runtime",
    )
    assert domain_success_runtime_suite.expectations[4].isEquivalentTo(
        exp6, match_type="success"
    )


def test_patch_expectation_add(exp5, exp7, domain_success_runtime_suite):
    assert domain_success_runtime_suite.expectations[4] is exp5

    assert not domain_success_runtime_suite.expectations[4].isEquivalentTo(
        exp7, match_type="success"
    )
    domain_success_runtime_suite.patch_expectation(
        exp5, op="add", path="/value_set/-", value=4, match_type="runtime",
    )
    assert domain_success_runtime_suite.expectations[4].isEquivalentTo(
        exp7, match_type="success"
    )


def test_patch_expectation_remove(exp5, exp8, domain_success_runtime_suite):
    assert domain_success_runtime_suite.expectations[4] is exp5

    assert not domain_success_runtime_suite.expectations[4].isEquivalentTo(
        exp8, match_type="runtime"
    )
    domain_success_runtime_suite.patch_expectation(
        exp5, op="remove", path="/result_format", value=None, match_type="runtime",
    )
    assert domain_success_runtime_suite.expectations[4].isEquivalentTo(
        exp8, match_type="runtime"
    )


def test_add_expectation(
    exp1,
    exp2,
    exp4,
    single_expectation_suite,
    baseline_suite,
    different_suite,
    domain_success_runtime_suite,
):
    assert len(single_expectation_suite.expectations) == 1
    assert not single_expectation_suite.isEquivalentTo(baseline_suite)
    single_expectation_suite.add_expectation(
        exp2, match_type="runtime", overwrite_existing=False
    )
    assert single_expectation_suite.isEquivalentTo(baseline_suite)
    assert len(single_expectation_suite.expectations) == 2

    # Should raise if overwrite_existing=False and a matching expectation is found
    with pytest.raises(DataContextError):
        single_expectation_suite.add_expectation(
            exp4, match_type="domain", overwrite_existing=False
        )

    assert not single_expectation_suite.isEquivalentTo(different_suite)
    single_expectation_suite.add_expectation(
        exp4, match_type="domain", overwrite_existing=True
    )
    assert single_expectation_suite.isEquivalentTo(different_suite)
    assert len(single_expectation_suite.expectations) == 2

    # Should raise if more than one matching expectation is found
    with pytest.raises(ValueError):
        domain_success_runtime_suite.add_expectation(
            exp2, match_type="success", overwrite_existing=False
        )

    with pytest.raises(InvalidExpectationConfigurationError):
        single_expectation_suite.add_expectation("not an expectation")

    # Turn this on once we're ready to enforce strict typing.
    # with pytest.raises(TypeError):
    #     single_expectation_suite.append_expectation(exp1.to_json_dict())


def test_remove_all_expectations_of_type(
    suite_with_table_and_column_expectations,
    suite_with_column_pair_and_table_expectations,
):
    assert not suite_with_table_and_column_expectations.isEquivalentTo(
        suite_with_column_pair_and_table_expectations
    )

    suite_with_table_and_column_expectations.remove_all_expectations_of_type(
        "expect_column_values_to_be_in_set"
    )

    assert suite_with_table_and_column_expectations.isEquivalentTo(
        suite_with_column_pair_and_table_expectations
    )



import pandas as pd
from tempfile import TemporaryDirectory
import great_expectations as ge
from great_expectations.dataset import PandasDataset
from great_expectations.data_context.data_context import ExpectationSuite
import os
import datetime


def test_save_expectation_suite_with_datetime_objects(data_context_parameterized_expectation_suite):
    # create datetime evaluation parameters
    evaluation_params = {"now": datetime.datetime.now(), "now_minus_48h": datetime.datetime.now() - datetime.timedelta(days=2)}
    test_data = {"data_refresh": [datetime.datetime.now(), datetime.datetime.now() - datetime.timedelta(days=1)]}
    test_df = pd.DataFrame(test_data)
    dataset_name = "test_pandas_source"

    with TemporaryDirectory() as tempdir:
        ge_path = os.path.join(tempdir, "great_expectations")
        ge.DataContext.create(tempdir, usage_statistics_enabled=False)
        context = ge.DataContext(ge_path)

        context.add_datasource(dataset_name, class_name="PandasDatasource")

        batch_kwargs = {"dataset": test_df, "datasource": dataset_name, "PandasInMemoryDF": True, "ge_batch_id": "test_id",}

        empty_suite = context.create_expectation_suite("test_suite")

        batch = context.get_batch(batch_kwargs=batch_kwargs, expectation_suite_name=empty_suite)
        for param in evaluation_params:
            batch.set_evaluation_parameter(param, evaluation_params[param])

        # Add expectation that will succeed using the datetime in a $PARAMETER
        batch.expect_column_max_to_be_between(column="data_refresh", min_value={"$PARAMETER": "now_minus_48h"})
        result = batch.validate()
        assert result.success
        batch.save_expectation_suite()
        assert isinstance(batch, PandasDataset)

        # Check that we can load the saved expectation suite
        reloaded_expectation_suite = context.get_expectation_suite("test_suite")
        assert isinstance(reloaded_expectation_suite, ExpectationSuite)

        # Run validation via the action_list_operator
        run_id = {
            "run_name": f"{dataset_name}_{datetime.datetime.now()}",
            "run_time": datetime.datetime.now(),
        }
        results = context.run_validation_operator(
            "action_list_operator", assets_to_validate=[batch], run_id=run_id,
            evaluation_parameters=evaluation_params
        )
        assert results.success

        # Check that we can build Data Docs
        index_page_locator_infos = context.build_data_docs()
        assert index_page_locator_infos["local_site"] == f"file://{ge_path}/uncommitted/data_docs/local_site/index.html"

        # Check that we can reload the expectation suite and validate
        reloaded_batch = context.get_batch(batch_kwargs=batch_kwargs, expectation_suite_name=reloaded_expectation_suite)

        run_id = {
            "run_name": f"reloaded_{dataset_name}_{datetime.datetime.now()}",
            "run_time": datetime.datetime.now(),
        }
        reloaded_results = context.run_validation_operator(
            "action_list_operator", assets_to_validate=[reloaded_batch], run_id=run_id,
        )

        assert reloaded_results.success



from copy import copy, deepcopy

import pytest

from great_expectations.core import ExpectationConfiguration, ExpectationSuite


@pytest.fixture
def exp1():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp2():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp3():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp4():
    return ExpectationConfiguration(
        expectation_type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def column_pair_expectation():
    return ExpectationConfiguration(
        expectation_type="expect_column_pair_values_to_be_in_set",
        kwargs={
            "column_A": "1",
            "column_B": "b",
            "value_set": [(1, 1), (2, 2)],
            "result_format": "BASIC",
        },
    )


@pytest.fixture
def table_exp1():
    return ExpectationConfiguration(
        expectation_type="expect_table_columns_to_match_ordered_list",
        kwargs={"value": ["a", "b", "c"]},
    )


@pytest.fixture
def table_exp2():
    return ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_be_between",
        kwargs={"min_value": 0, "max_value": 1},
    )


@pytest.fixture
def table_exp3():
    return ExpectationConfiguration(
        expectation_type="expect_table_row_count_to_equal", kwargs={"value": 1}
    )


@pytest.fixture
def empty_suite():
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def single_expectation_suite(exp1):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def suite_with_table_and_column_expectations(
    exp1, exp2, exp3, exp4, column_pair_expectation, table_exp1, table_exp2, table_exp3
):
    suite = ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[
            exp1,
            exp2,
            exp3,
            exp4,
            column_pair_expectation,
            table_exp1,
            table_exp2,
            table_exp3,
        ],
        meta={"notes": "This is an expectation suite."},
    )
    assert suite.expectations == [
        exp1,
        exp2,
        exp3,
        exp4,
        column_pair_expectation,
        table_exp1,
        table_exp2,
        table_exp3,
    ]
    return suite


@pytest.fixture
def baseline_suite(exp1, exp2):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1, exp2],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def identical_suite(exp1, exp3):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1, exp3],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def equivalent_suite(exp1, exp3):
    return ExpectationSuite(
        expectation_suite_name="danger",
        expectations=[exp1, exp3],
        meta={
            "notes": "This is another expectation suite, with a different name and meta"
        },
    )


@pytest.fixture
def different_suite(exp1, exp4):
    return ExpectationSuite(
        expectation_suite_name="warning",
        expectations=[exp1, exp4],
        meta={"notes": "This is an expectation suite."},
    )


def test_expectation_suite_equality(baseline_suite, identical_suite, equivalent_suite):
    """Equality should depend on all defined properties of a configuration object, but not on whether the *instances*
    are the same."""
    assert baseline_suite is baseline_suite  # no difference
    assert (
        baseline_suite is not identical_suite
    )  # different instances, but same content
    assert baseline_suite == identical_suite  # different instances, but same content
    assert not (baseline_suite != identical_suite)  # ne works properly
    assert not (baseline_suite == equivalent_suite)  # different meta
    assert baseline_suite != equivalent_suite  # ne works properly


def test_expectation_suite_equivalence(
    baseline_suite,
    identical_suite,
    equivalent_suite,
    different_suite,
    single_expectation_suite,
):
    """Equivalence should depend only on properties that affect the result of the expectation."""
    assert baseline_suite.isEquivalentTo(baseline_suite)  # no difference
    assert baseline_suite.isEquivalentTo(identical_suite)
    assert baseline_suite.isEquivalentTo(equivalent_suite)  # different meta
    assert not baseline_suite.isEquivalentTo(
        different_suite
    )  # different value_set in one expectation
    assert not single_expectation_suite.isEquivalentTo(baseline_suite)


def test_expectation_suite_dictionary_equivalence(baseline_suite):
    assert (
        baseline_suite.isEquivalentTo(
            {
                "expectation_suite_name": "warning",
                "expectations": [
                    {
                        "expectation_type": "expect_column_values_to_be_in_set",
                        "kwargs": {
                            "column": "a",
                            "value_set": [1, 2, 3],
                            "result_format": "BASIC",
                        },
                        "meta": {"notes": "This is an expectation."},
                    },
                    {
                        "expectation_type": "expect_column_values_to_be_in_set",
                        "kwargs": {
                            "column": "b",
                            "value_set": [-1, -2, -3],
                            "result_format": "BASIC",
                        },
                        "meta": {"notes": "This is an expectation."},
                    },
                ],
                "meta": {"notes": "This is an expectation suite."},
            }
        )
        is True
    )

    assert (
        baseline_suite.isEquivalentTo(
            {
                "expectation_suite_name": "warning",
                "expectations": [
                    {
                        "expectation_type": "expect_column_values_to_be_in_set",
                        "kwargs": {
                            "column": "a",
                            "value_set": [-1, 2, 3],  # One value changed here
                            "result_format": "BASIC",
                        },
                        "meta": {"notes": "This is an expectation."},
                    },
                    {
                        "expectation_type": "expect_column_values_to_be_in_set",
                        "kwargs": {
                            "column": "b",
                            "value_set": [-1, -2, -3],
                            "result_format": "BASIC",
                        },
                        "meta": {"notes": "This is an expectation."},
                    },
                ],
                "meta": {"notes": "This is an expectation suite."},
            }
        )
        is False
    )


def test_expectation_suite_copy(baseline_suite):
    suite_copy = copy(baseline_suite)
    assert suite_copy == baseline_suite
    suite_copy.data_asset_type = "blarg!"
    assert (
        baseline_suite.data_asset_type != "blarg"
    )  # copy on primitive properties shouldn't propagate
    suite_copy.expectations[0].meta["notes"] = "a different note"
    assert (
        baseline_suite.expectations[0].meta["notes"] == "a different note"
    )  # copy on deep attributes does propagate


def test_expectation_suite_deepcopy(baseline_suite):
    suite_deepcopy = deepcopy(baseline_suite)
    assert suite_deepcopy == baseline_suite
    suite_deepcopy.data_asset_type = "blarg!"
    assert (
        baseline_suite.data_asset_type != "blarg"
    )  # copy on primitive properties shouldn't propagate
    suite_deepcopy.expectations[0].meta["notes"] = "a different note"
    # deepcopy on deep attributes does not propagate
    assert baseline_suite.expectations[0].meta["notes"] == "This is an expectation."


def test_suite_without_metadata_includes_ge_version_metadata_if_none_is_provided():
    suite = ExpectationSuite("foo")
    assert "great_expectations_version" in suite.meta.keys()


def test_suite_does_not_overwrite_existing_version_metadata():
    suite = ExpectationSuite("foo", meta={"great_expectations_version": "0.0.0"})
    assert "great_expectations_version" in suite.meta.keys()
    assert suite.meta["great_expectations_version"] == "0.0.0"


def test_suite_with_metadata_includes_ge_version_metadata(baseline_suite):
    assert "great_expectations_version" in baseline_suite.meta.keys()


def test_add_citation(baseline_suite):
    assert (
        "citations" not in baseline_suite.meta
        or len(baseline_suite.meta["citations"]) == 0
    )
    baseline_suite.add_citation("hello!")
    assert baseline_suite.meta["citations"][0].get("comment") == "hello!"


def test_get_citations_with_no_citations(baseline_suite):
    assert "citations" not in baseline_suite.meta
    assert baseline_suite.get_citations() == []


def test_get_citations_not_sorted(baseline_suite):
    assert "citations" not in baseline_suite.meta
    baseline_suite.add_citation("first", citation_date="2000-01-01")
    baseline_suite.add_citation("third", citation_date="2000-01-03")
    baseline_suite.add_citation("second", citation_date="2000-01-02")
    assert baseline_suite.get_citations(sort=False) == [
        {
            "batch_kwargs": None,
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-01",
            "comment": "first",
        },
        {
            "batch_kwargs": None,
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-03",
            "comment": "third",
        },
        {
            "batch_kwargs": None,
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-02",
            "comment": "second",
        },
    ]


def test_get_citations_sorted(baseline_suite):
    assert "citations" not in baseline_suite.meta
    baseline_suite.add_citation("first", citation_date="2000-01-01")
    baseline_suite.add_citation("third", citation_date="2000-01-03")
    baseline_suite.add_citation("second", citation_date="2000-01-02")
    assert baseline_suite.get_citations(sort=True) == [
        {
            "batch_kwargs": None,
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-01",
            "comment": "first",
        },
        {
            "batch_kwargs": None,
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-02",
            "comment": "second",
        },
        {
            "batch_kwargs": None,
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-03",
            "comment": "third",
        },
    ]


def test_get_citations_with_multiple_citations_containing_batch_kwargs(baseline_suite):
    assert "citations" not in baseline_suite.meta
    baseline_suite.add_citation(
        "first", batch_kwargs={"path": "first"}, citation_date="2000-01-01"
    )
    baseline_suite.add_citation(
        "second", batch_kwargs={"path": "second"}, citation_date="2001-01-01"
    )
    baseline_suite.add_citation("third", citation_date="2002-01-01")

    obs = baseline_suite.get_citations(sort=True, require_batch_kwargs=True)
    assert obs == [
        {
            "batch_kwargs": {"path": "first"},
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2000-01-01",
            "comment": "first",
        },
        {
            "batch_kwargs": {"path": "second"},
            "batch_markers": None,
            "batch_parameters": None,
            "citation_date": "2001-01-01",
            "comment": "second",
        },
    ]


def test_get_table_expectations_returns_empty_list_on_empty_suite(empty_suite):
    assert empty_suite.get_table_expectations() == []


def test_get_table_expectations_returns_empty_list_on_suite_without_any(baseline_suite):
    assert baseline_suite.get_table_expectations() == []


def test_get_table_expectations(
    suite_with_table_and_column_expectations, table_exp1, table_exp2, table_exp3
):
    obs = suite_with_table_and_column_expectations.get_table_expectations()
    assert obs == [table_exp1, table_exp2, table_exp3]


def test_get_column_expectations_returns_empty_list_on_empty_suite(empty_suite):
    assert empty_suite.get_column_expectations() == []


def test_get_column_expectations(
    suite_with_table_and_column_expectations, exp1, exp2, exp3, exp4
):
    obs = suite_with_table_and_column_expectations.get_column_expectations()
    assert obs == [exp1, exp2, exp3, exp4]




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2020-10-20T11:38:21-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "843d3d3d8f01fc4d8ab2eaf0f42060abfa32e405",
 "version": "0.12.6"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""
Test the ability to use datetime objects as run time evaluation parameters.
"""

import pandas as pd

from great_expectations.dataset import PandasDataset


def test_pandas_datetime_evaluation_parameter():
    evaluation_params = {"now": pd.Timestamp.now(), "now_minus_48h": pd.Timestamp.now() - pd.to_timedelta(2, unit="d")}

    test_data = {
        "data_refresh": [pd.Timestamp.now(), (pd.Timestamp.now() - pd.to_timedelta(1, unit="d"))]
    }
    _df = pd.DataFrame(test_data)
    df = PandasDataset(_df)

    for param in evaluation_params:
        df.set_evaluation_parameter(param, evaluation_params[param])
    df.expect_column_max_to_be_between(column="data_refresh", min_value={"$PARAMETER": "now_minus_48h"})

    result = df.validate()

    assert result.success



from setuptools import find_packages, setup

import versioneer

# Parse requirements.txt
with open("requirements.txt") as f:
    required = f.read().splitlines()

# try:
#    import pypandoc
#    long_description = pypandoc.convert_file('README.md', 'rst')
# except (IOError, ImportError):
long_description = "Always know what to expect from your data. (See https://github.com/great-expectations/great_expectations for full description)."

config = {
    "description": "Always know what to expect from your data.",
    "author": "The Great Expectations Team",
    "url": "https://github.com/great-expectations/great_expectations",
    "author_email": "team@greatexpectations.io",
    "version": versioneer.get_version(),
    "cmdclass": versioneer.get_cmdclass(),
    "install_requires": required,
    "extras_require": {
        "spark": ["pyspark>=2.3.2"],
        "sqlalchemy": ["sqlalchemy>=1.2"],
        "airflow": ["apache-airflow[s3]>=1.9.0", "boto3>=1.7.3"],
        "gcp": [
            "google-cloud>=0.34.0",
            "google-cloud-storage>=1.28.0",
            "google-cloud-secret-manager>=1.0.0",
            "pybigquery==0.4.15",
        ],
        "redshift": ["psycopg2>=2.8"],
        "s3": ["boto3>=1.14"],
        "snowflake": ["snowflake-sqlalchemy>=1.2"],
    },
    "packages": find_packages(exclude=["docs*", "tests*", "examples*"]),
    "entry_points": {
        "console_scripts": ["great_expectations=great_expectations.cli:main"]
    },
    "name": "great_expectations",
    "long_description": long_description,
    "license": "Apache-2.0",
    "keywords": "data science testing pipeline data quality dataquality validation datavalidation",
    "include_package_data": True,
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Other Audience",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Testing",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
}

setup(**config)



import json
import os
import shutil

from click.testing import CliRunner
from freezegun import freeze_time

import great_expectations
from great_expectations import DataContext
from great_expectations.cli import cli
from great_expectations.data_context.util import file_relative_path
from great_expectations.util import gen_directory_tree_str
from tests.cli.utils import assert_no_logging_messages_or_tracebacks

try:
    from unittest import mock
except ImportError:
    from unittest import mock


def test_project_upgrade_already_up_to_date(v10_project_directory, caplog):
    # test great_expectations project upgrade command with project with config_version 2

    # copy v2 yml
    shutil.copy(
        file_relative_path(
            __file__, "../../test_fixtures/upgrade_helper/great_expectations_v2.yml"
        ),
        os.path.join(v10_project_directory, "great_expectations.yml"),
    )

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli,
        ["project", "upgrade", "-d", v10_project_directory],
        input="\n",
        catch_exceptions=False,
    )
    stdout = result.stdout

    assert "Checking project..." in stdout
    assert "Your project is up-to-date - no upgrade is necessary." in stdout
    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_upgrade_helper_intervention_on_cli_command(v10_project_directory, caplog):
    # test if cli detects out of date project and asks to run upgrade helper
    # decline upgrade and ensure config version was not modified

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli,
        ["suite", "list", "-d", v10_project_directory],
        input="n\n",
        catch_exceptions=False,
    )
    stdout = result.stdout

    assert (
        "Your project appears to have an out-of-date config version (1.0) - the version number must be at least 2."
        in stdout
    )
    assert "In order to proceed, your project must be upgraded." in stdout
    assert (
        "Would you like to run the Upgrade Helper to bring your project up-to-date? [Y/n]:"
        in stdout
    )
    assert (
        "Ok, exiting now. To upgrade at a later time, use the following command: [36mgreat_expectations project "
        "upgrade[0m" in stdout
    )
    assert (
        "To learn more about the upgrade process, visit ["
        "36mhttps://docs.greatexpectations.io/en/latest/how_to_guides/migrating_versions.html"
        in stdout
    )
    assert_no_logging_messages_or_tracebacks(caplog, result)

    # make sure config version unchanged
    assert (
        DataContext.get_ge_config_version(context_root_dir=v10_project_directory) == 1
    )

    expected_project_tree_str = """\
great_expectations/
    .gitignore
    great_expectations.yml
    checkpoints/
        .gitkeep
    expectations/
        .gitkeep
    notebooks/
        .gitkeep
    plugins/
        custom_store_backends/
            __init__.py
            my_custom_store_backend.py
    uncommitted/
        config_variables.yml
        data_docs/
            local_site/
                expectations/
                    .gitkeep
                static/
                    .gitkeep
                validations/
                    diabetic_data/
                        warning/
                            20200430T191246.763896Z/
                                c3b4c5df224fef4b1a056a0f3b93aba5.html
        validations/
            diabetic_data/
                warning/
                    20200430T191246.763896Z/
                        c3b4c5df224fef4b1a056a0f3b93aba5.json
"""
    obs_project_tree_str = gen_directory_tree_str(v10_project_directory)
    assert obs_project_tree_str == expected_project_tree_str


@freeze_time("09/26/2019 13:42:41")
def test_basic_project_upgrade(v10_project_directory, caplog):
    # test project upgrade that requires no manual steps

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli,
        ["project", "upgrade", "-d", v10_project_directory],
        input="\n",
        catch_exceptions=False,
    )
    stdout = result.stdout

    with open(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/test_basic_project_upgrade_expected_stdout.fixture",
        )
    ) as f:
        expected_stdout = f.read()
        expected_stdout = expected_stdout.replace(
            "GE_PROJECT_DIR", v10_project_directory
        )
        assert stdout == expected_stdout

    expected_project_tree_str = """\
great_expectations/
    .gitignore
    great_expectations.yml
    checkpoints/
        .gitkeep
    expectations/
        .gitkeep
    notebooks/
        .gitkeep
    plugins/
        custom_store_backends/
            __init__.py
            my_custom_store_backend.py
    uncommitted/
        config_variables.yml
        data_docs/
            local_site/
                expectations/
                    .gitkeep
                static/
                    .gitkeep
                validations/
                    diabetic_data/
                        warning/
                            20200430T191246.763896Z/
                                20200430T191246.763896Z/
                                    c3b4c5df224fef4b1a056a0f3b93aba5.html
        logs/
            project_upgrades/
                UpgradeHelperV11_20190926T134241.000000Z.json
        validations/
            diabetic_data/
                warning/
                    20200430T191246.763896Z/
                        20200430T191246.763896Z/
                            c3b4c5df224fef4b1a056a0f3b93aba5.json
"""
    obs_project_tree_str = gen_directory_tree_str(v10_project_directory)
    assert obs_project_tree_str == expected_project_tree_str
    # make sure config number incremented
    assert (
        DataContext.get_ge_config_version(context_root_dir=v10_project_directory) == 2
    )

    with open(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/UpgradeHelperV11_basic_upgrade_log.json",
        )
    ) as f:
        expected_upgrade_log_dict = json.load(f)
        expected_upgrade_log_str = json.dumps(expected_upgrade_log_dict)
        expected_upgrade_log_str = expected_upgrade_log_str.replace(
            "GE_PROJECT_DIR", v10_project_directory
        )
        expected_upgrade_log_dict = json.loads(expected_upgrade_log_str)

    with open(
        f"{v10_project_directory}/uncommitted/logs/project_upgrades/UpgradeHelperV11_20190926T134241.000000Z"
        f".json"
    ) as f:
        obs_upgrade_log_dict = json.load(f)

    assert obs_upgrade_log_dict == expected_upgrade_log_dict


@freeze_time("09/26/2019 13:42:41")
def test_project_upgrade_with_manual_steps(
    v10_project_directory, caplog, sa, postgresql_engine
):
    # This test requires sqlalchemy because it includes database backends configured
    # test project upgrade that requires manual steps

    # copy v2 yml
    shutil.copy(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/great_expectations_v1_needs_manual_upgrade"
            ".yml",
        ),
        os.path.join(v10_project_directory, "great_expectations.yml"),
    )

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli,
        ["project", "upgrade", "-d", v10_project_directory],
        input="\n",
        catch_exceptions=False,
    )
    stdout = result.stdout

    with open(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/test_project_upgrade_with_manual_steps_expected_stdout.fixture",
        )
    ) as f:
        expected_stdout = f.read()
        expected_stdout = expected_stdout.replace(
            "GE_PROJECT_DIR", v10_project_directory
        )
        assert stdout == expected_stdout

    expected_project_tree_str = """\
great_expectations/
    .gitignore
    great_expectations.yml
    checkpoints/
        .gitkeep
    expectations/
        .gitkeep
    notebooks/
        .gitkeep
    plugins/
        custom_store_backends/
            __init__.py
            my_custom_store_backend.py
    uncommitted/
        config_variables.yml
        data_docs/
            local_site/
                expectations/
                    .gitkeep
                static/
                    .gitkeep
                validations/
                    diabetic_data/
                        warning/
                            20200430T191246.763896Z/
                                20200430T191246.763896Z/
                                    c3b4c5df224fef4b1a056a0f3b93aba5.html
        logs/
            project_upgrades/
                UpgradeHelperV11_20190926T134241.000000Z.json
        validations/
            diabetic_data/
                warning/
                    20200430T191246.763896Z/
                        20200430T191246.763896Z/
                            c3b4c5df224fef4b1a056a0f3b93aba5.json
"""
    obs_project_tree_str = gen_directory_tree_str(v10_project_directory)
    assert obs_project_tree_str == expected_project_tree_str
    # make sure config number not incremented
    assert (
        DataContext.get_ge_config_version(context_root_dir=v10_project_directory) == 1
    )

    with open(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/UpgradeHelperV11_manual_steps_upgrade_log.json",
        )
    ) as f:
        expected_upgrade_log_dict = json.load(f)
        expected_upgrade_log_str = json.dumps(expected_upgrade_log_dict)
        expected_upgrade_log_str = expected_upgrade_log_str.replace(
            "GE_PROJECT_DIR", v10_project_directory
        )
        expected_upgrade_log_dict = json.loads(expected_upgrade_log_str)

    with open(
        f"{v10_project_directory}/uncommitted/logs/project_upgrades/UpgradeHelperV11_20190926T134241.000000Z"
        f".json"
    ) as f:
        obs_upgrade_log_dict = json.load(f)

    assert obs_upgrade_log_dict == expected_upgrade_log_dict


@freeze_time("09/26/2019 13:42:41")
def test_project_upgrade_with_exception(v10_project_directory, caplog):
    # test project upgrade that requires manual steps

    # copy v2 yml
    shutil.copy(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/great_expectations_v1_basic_with_exception"
            ".yml",
        ),
        os.path.join(v10_project_directory, "great_expectations.yml"),
    )

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli,
        ["project", "upgrade", "-d", v10_project_directory],
        input="\n",
        catch_exceptions=False,
    )
    stdout = result.stdout

    with open(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/test_project_upgrade_with_exception_expected_stdout.fixture",
        )
    ) as f:
        expected_stdout = f.read()
        expected_stdout = expected_stdout.replace(
            "GE_PROJECT_DIR", v10_project_directory
        )
        assert stdout == expected_stdout

    expected_project_tree_str = """\
great_expectations/
    .gitignore
    great_expectations.yml
    checkpoints/
        .gitkeep
    expectations/
        .gitkeep
    notebooks/
        .gitkeep
    plugins/
        custom_store_backends/
            __init__.py
            my_custom_store_backend.py
    uncommitted/
        config_variables.yml
        data_docs/
            local_site/
                expectations/
                    .gitkeep
                static/
                    .gitkeep
                validations/
                    diabetic_data/
                        warning/
                            20200430T191246.763896Z/
                                20200430T191246.763896Z/
                                    c3b4c5df224fef4b1a056a0f3b93aba5.html
        logs/
            project_upgrades/
                UpgradeHelperV11_20190926T134241.000000Z.json
        validations/
            diabetic_data/
                warning/
                    20200430T191246.763896Z/
                        20200430T191246.763896Z/
                            c3b4c5df224fef4b1a056a0f3b93aba5.json
"""
    obs_project_tree_str = gen_directory_tree_str(v10_project_directory)
    assert obs_project_tree_str == expected_project_tree_str
    # make sure config number not incremented
    assert (
        DataContext.get_ge_config_version(context_root_dir=v10_project_directory) == 1
    )

    with open(
        file_relative_path(
            __file__,
            "../../test_fixtures/upgrade_helper/UpgradeHelperV11_basic_upgrade_with_exception_log"
            ".json",
        )
    ) as f:
        expected_upgrade_log_dict = json.load(f)
        expected_upgrade_log_str = json.dumps(expected_upgrade_log_dict)
        expected_upgrade_log_str = expected_upgrade_log_str.replace(
            "GE_PROJECT_DIR", v10_project_directory
        )
        expected_upgrade_log_str = expected_upgrade_log_str.replace(
            "GE_PATH", os.path.split(great_expectations.__file__)[0]
        )
        expected_upgrade_log_dict = json.loads(expected_upgrade_log_str)

    with open(
        f"{v10_project_directory}/uncommitted/logs/project_upgrades/UpgradeHelperV11_20190926T134241.000000Z"
        f".json"
    ) as f:
        obs_upgrade_log_dict = json.load(f)
        obs_upgrade_log_dict["exceptions"][0]["exception_message"] = ""

    assert obs_upgrade_log_dict == expected_upgrade_log_dict



import datetime
import json

import pandas as pd
import pytest

import great_expectations as ge
from great_expectations.core import ExpectationConfiguration, expectationSuiteSchema
from great_expectations.profile import ColumnsExistProfiler
from tests.test_utils import expectationValidationResultSchema


def test_expect_column_values_to_be_dateutil_parseable():
    D = ge.dataset.PandasDataset(
        {
            "c1": ["03/06/09", "23 April 1973", "January 9, 2016"],
            "c2": ["9/8/2012", "covfefe", 25],
            "c3": ["Jared", "June 1, 2013", "July 18, 1976"],
            "c4": ["1", "2", "49000004632"],
            "already_datetime": [
                datetime.datetime(2015, 1, 1),
                datetime.datetime(2016, 1, 1),
                datetime.datetime(2017, 1, 1),
            ],
        }
    )
    D.set_default_expectation_argument("result_format", "COMPLETE")

    T = [
        {
            "in": {"column": "c1"},
            "out": {
                "success": True,
                "unexpected_list": [],
                "unexpected_index_list": [],
            },
        },
        {
            "in": {"column": "c2", "catch_exceptions": True},
            # 'out':{'success':False, 'unexpected_list':['covfefe', 25], 'unexpected_index_list': [1, 2]}},
            "error": {
                "traceback_substring": "TypeError: Values passed to expect_column_values_to_be_dateutil_parseable must be of type string"
            },
        },
        {
            "in": {"column": "c3"},
            "out": {
                "success": False,
                "unexpected_list": ["Jared"],
                "unexpected_index_list": [0],
            },
        },
        {
            "in": {"column": "c3", "mostly": 0.5},
            "out": {
                "success": True,
                "unexpected_list": ["Jared"],
                "unexpected_index_list": [0],
            },
        },
        {
            "in": {"column": "c4"},
            "out": {
                "success": False,
                "unexpected_list": ["49000004632"],
                "unexpected_index_list": [2],
            },
        },
        {
            "in": {"column": "already_datetime", "catch_exceptions": True},
            "error": {
                "traceback_substring": "TypeError: Values passed to expect_column_values_to_be_dateutil_parseable must be of type string"
            },
        },
    ]

    for t in T:
        out = D.expect_column_values_to_be_dateutil_parseable(**t["in"])
        if "out" in t:
            assert t["out"]["success"] == out.success
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]
        elif "error" in t:
            assert out.exception_info["raised_exception"] is True
            assert (
                t["error"]["traceback_substring"]
                in out.exception_info["exception_traceback"]
            )


def test_expect_column_values_to_be_json_parseable():
    d1 = json.dumps({"i": [1, 2, 3], "j": 35, "k": {"x": "five", "y": 5, "z": "101"}})
    d2 = json.dumps({"i": 1, "j": 2, "k": [3, 4, 5]})
    d3 = json.dumps({"i": "a", "j": "b", "k": "c"})
    d4 = json.dumps(
        {"i": [4, 5], "j": [6, 7], "k": [8, 9], "l": {4: "x", 5: "y", 6: "z"}}
    )
    D = ge.dataset.PandasDataset(
        {
            "json_col": [d1, d2, d3, d4],
            "not_json": [4, 5, 6, 7],
            "py_dict": [
                {"a": 1, "out": 1},
                {"b": 2, "out": 4},
                {"c": 3, "out": 9},
                {"d": 4, "out": 16},
            ],
            "most": [d1, d2, d3, "d4"],
        }
    )
    D.set_default_expectation_argument("result_format", "COMPLETE")

    T = [
        {
            "in": {"column": "json_col"},
            "out": {
                "success": True,
                "unexpected_index_list": [],
                "unexpected_list": [],
            },
        },
        {
            "in": {"column": "not_json"},
            "out": {
                "success": False,
                "unexpected_index_list": [0, 1, 2, 3],
                "unexpected_list": [4, 5, 6, 7],
            },
        },
        {
            "in": {"column": "py_dict"},
            "out": {
                "success": False,
                "unexpected_index_list": [0, 1, 2, 3],
                "unexpected_list": [
                    {"a": 1, "out": 1},
                    {"b": 2, "out": 4},
                    {"c": 3, "out": 9},
                    {"d": 4, "out": 16},
                ],
            },
        },
        {
            "in": {"column": "most"},
            "out": {
                "success": False,
                "unexpected_index_list": [3],
                "unexpected_list": ["d4"],
            },
        },
        {
            "in": {"column": "most", "mostly": 0.75},
            "out": {
                "success": True,
                "unexpected_index_list": [3],
                "unexpected_list": ["d4"],
            },
        },
    ]

    for t in T:
        out = D.expect_column_values_to_be_json_parseable(**t["in"])
        assert t["out"]["success"] == out.success
        assert t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
        assert t["out"]["unexpected_list"] == out.result["unexpected_list"]


def test_expectation_decorator_summary_mode():
    df = ge.dataset.PandasDataset({"x": [1, 2, 3, 4, 5, 6, 7, 7, None, None],})
    df.set_default_expectation_argument("result_format", "COMPLETE")
    df.set_default_expectation_argument("include_config", False)

    # print '&'*80
    # print json.dumps(df.expect_column_values_to_be_between('x', min_value=1, max_value=5, result_format="SUMMARY"), indent=2)

    exp_output = expectationValidationResultSchema.load(
        {
            "success": False,
            "result": {
                "element_count": 10,
                "missing_count": 2,
                "missing_percent": 20.0,
                "unexpected_count": 3,
                "partial_unexpected_counts": [
                    {"value": 7.0, "count": 2},
                    {"value": 6.0, "count": 1},
                ],
                "unexpected_percent": 30.0,
                "unexpected_percent_nonmissing": 37.5,
                "partial_unexpected_list": [6.0, 7.0, 7.0],
                "partial_unexpected_index_list": [5, 6, 7],
            },
        }
    )
    assert (
        df.expect_column_values_to_be_between(
            "x", min_value=1, max_value=5, result_format="SUMMARY"
        )
        == exp_output
    )

    exp_output = expectationValidationResultSchema.load(
        {
            "success": True,
            "result": {
                "observed_value": 4.375,
                "element_count": 10,
                "missing_count": 2,
                "missing_percent": 20.0,
            },
        }
    )

    assert (
        df.expect_column_mean_to_be_between("x", 3, 7, result_format="SUMMARY")
        == exp_output
    )


def test_positional_arguments():

    df = ge.dataset.PandasDataset(
        {"x": [1, 3, 5, 7, 9], "y": [2, 4, 6, 8, 10], "z": [None, "a", "b", "c", "abc"]}
    )
    df.set_default_expectation_argument("result_format", "COMPLETE")
    df.set_default_expectation_argument("include_config", False)

    exp_output = expectationValidationResultSchema.load(
        {
            "success": True,
            "result": {
                "observed_value": 5,
                "element_count": 5,
                "missing_count": 0,
                "missing_percent": 0.0,
            },
        }
    )

    assert df.expect_column_mean_to_be_between("x", 4, 6) == exp_output

    out = df.expect_column_values_to_be_between("y", 1, 6)
    t = {
        "out": {
            "success": False,
            "unexpected_list": [8, 10],
            "unexpected_index_list": [3, 4],
        }
    }
    if "out" in t:
        assert t["out"]["success"] == out.success
        if "unexpected_index_list" in t["out"]:
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
        if "unexpected_list" in t["out"]:
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]

    out = df.expect_column_values_to_be_between("y", 1, 8, strict_max=True)
    t = {
        "out": {
            "success": False,
            "unexpected_list": [8, 10],
            "unexpected_index_list": [3, 4],
        }
    }
    if "out" in t:
        assert t["out"]["success"] == out.success
        if "unexpected_index_list" in t["out"]:
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
        if "unexpected_list" in t["out"]:
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]

    out = df.expect_column_values_to_be_between("y", 2, 100, strict_min=True)
    t = {
        "out": {"success": False, "unexpected_list": [2], "unexpected_index_list": [0]}
    }
    if "out" in t:
        assert t["out"]["success"] == out.success
        if "unexpected_index_list" in t["out"]:
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
        if "unexpected_list" in t["out"]:
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]

    out = df.expect_column_values_to_be_between("y", 1, 6, mostly=0.5)
    t = {
        "out": {
            "success": True,
            "unexpected_list": [8, 10],
            "unexpected_index_list": [3, 4],
        }
    }
    if "out" in t:
        assert t["out"]["success"] == out.success
        if "unexpected_index_list" in t["out"]:
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
        if "unexpected_list" in t["out"]:
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]

    out = df.expect_column_values_to_be_in_set("z", ["a", "b", "c"])
    t = {
        "out": {
            "success": False,
            "unexpected_list": ["abc"],
            "unexpected_index_list": [4],
        }
    }
    if "out" in t:
        assert t["out"]["success"] == out.success
        if "unexpected_index_list" in t["out"]:
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
        if "unexpected_list" in t["out"]:
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]

    out = df.expect_column_values_to_be_in_set("z", ["a", "b", "c"], mostly=0.5)
    t = {
        "out": {
            "success": True,
            "unexpected_list": ["abc"],
            "unexpected_index_list": [4],
        }
    }
    if "out" in t:
        assert t["out"]["success"] == out.success
        if "unexpected_index_list" in t["out"]:
            assert (
                t["out"]["unexpected_index_list"] == out.result["unexpected_index_list"]
            )
        if "unexpected_list" in t["out"]:
            assert t["out"]["unexpected_list"] == out.result["unexpected_list"]


def test_result_format_argument_in_decorators():
    df = ge.dataset.PandasDataset(
        {"x": [1, 3, 5, 7, 9], "y": [2, 4, 6, 8, 10], "z": [None, "a", "b", "c", "abc"]}
    )
    df.set_default_expectation_argument("result_format", "COMPLETE")
    df.set_default_expectation_argument("include_config", False)

    # Test explicit Nones in result_format
    exp_output = expectationValidationResultSchema.load(
        {
            "success": True,
            "result": {
                "observed_value": 5,
                "element_count": 5,
                "missing_count": 0,
                "missing_percent": 0.0,
            },
        }
    )
    assert (
        df.expect_column_mean_to_be_between("x", 4, 6, result_format=None) == exp_output
    )

    exp_output = expectationValidationResultSchema.load(
        {
            "result": {
                "element_count": 5,
                "missing_count": 0,
                "missing_percent": 0.0,
                "partial_unexpected_counts": [
                    {"count": 1, "value": 8},
                    {"count": 1, "value": 10},
                ],
                "partial_unexpected_index_list": [3, 4],
                "partial_unexpected_list": [8, 10],
                "unexpected_count": 2,
                "unexpected_index_list": [3, 4],
                "unexpected_list": [8, 10],
                "unexpected_percent": 40.0,
                "unexpected_percent_nonmissing": 40.0,
            },
            "success": False,
        }
    )

    assert (
        df.expect_column_values_to_be_between("y", 1, 6, result_format=None)
        == exp_output
    )

    # Test unknown output format
    with pytest.raises(ValueError):
        df.expect_column_values_to_be_between("y", 1, 6, result_format="QUACK")

    with pytest.raises(ValueError):
        df.expect_column_mean_to_be_between("x", 4, 6, result_format="QUACK")


def test_from_pandas():
    pd_df = pd.DataFrame(
        {"x": [1, 3, 5, 7, 9], "y": [2, 4, 6, 8, 10], "z": [None, "a", "b", "c", "abc"]}
    )

    ge_df = ge.from_pandas(pd_df)
    assert isinstance(ge_df, ge.data_asset.DataAsset)
    assert list(ge_df.columns) == ["x", "y", "z"]
    assert list(ge_df["x"]) == list(pd_df["x"])
    assert list(ge_df["y"]) == list(pd_df["y"])
    assert list(ge_df["z"]) == list(pd_df["z"])

    # make an empty subclass to test dataset_class argument
    class CustomPandasDataset(ge.dataset.PandasDataset):
        pass

    ge_df_custom = ge.from_pandas(pd_df, dataset_class=CustomPandasDataset)

    assert not isinstance(ge_df, CustomPandasDataset)
    assert isinstance(ge_df_custom, CustomPandasDataset)
    assert list(ge_df_custom.columns) == ["x", "y", "z"]
    assert list(ge_df_custom["x"]) == list(pd_df["x"])
    assert list(ge_df_custom["y"]) == list(pd_df["y"])
    assert list(ge_df_custom["z"]) == list(pd_df["z"])


def test_ge_pandas_concatenating_no_autoinspect():
    df1 = ge.dataset.PandasDataset({"A": ["A0", "A1", "A2"], "B": ["B0", "B1", "B2"]})

    df1.expect_column_to_exist("A")
    df1.expect_column_to_exist("B")
    df1.expect_column_values_to_match_regex("A", "^A[0-2]$")
    df1.expect_column_values_to_match_regex("B", "^B[0-2]$")

    df2 = ge.dataset.PandasDataset({"A": ["A3", "A4", "A5"], "B": ["B3", "B4", "B5"]})

    df2.expect_column_to_exist("A")
    df2.expect_column_to_exist("B")
    df2.expect_column_values_to_match_regex("A", "^A[3-5]$")
    df2.expect_column_values_to_match_regex("B", "^B[3-5]$")

    df = pd.concat([df1, df2])

    exp_c = []

    # The concatenated data frame will:
    #
    #   1. Be a ge.dataset.PandaDataSet
    #   2. Have no expectations (since no default expectations are created), even expectations that were common
    #      to the concatenated dataframes and still make sense (since no autoinspection happens).

    assert isinstance(df, ge.dataset.PandasDataset)
    assert df.get_expectation_suite().expectations == exp_c


def test_ge_pandas_joining():
    df1 = ge.dataset.PandasDataset(
        {"A": ["A0", "A1", "A2"], "B": ["B0", "B1", "B2"]}, index=["K0", "K1", "K2"]
    )

    df1.expect_column_values_to_match_regex("A", "^A[0-2]$")
    df1.expect_column_values_to_match_regex("B", "^B[0-2]$")

    df2 = ge.dataset.PandasDataset(
        {"C": ["C0", "C2", "C3"], "D": ["C0", "D2", "D3"]}, index=["K0", "K2", "K3"]
    )

    df2.expect_column_values_to_match_regex("C", "^C[0-2]$")
    df2.expect_column_values_to_match_regex("D", "^D[0-2]$")

    df = df1.join(df2)

    exp_j = [
        # No autoinspection is default 20180920
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'A'}},
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'B'}},
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'C'}},
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'D'}}
    ]

    # The joined data frame will:
    #
    #   1. Be a ge.dataset.PandaDataSet
    #   2. Have no expectations (no autoinspection)

    assert isinstance(df, ge.dataset.PandasDataset)
    assert df.get_expectation_suite().expectations == exp_j


def test_ge_pandas_merging():
    df1 = ge.dataset.PandasDataset({"id": [1, 2, 3, 4], "name": ["a", "b", "c", "d"]})

    df1.expect_column_values_to_match_regex("name", "^[A-Za-z ]+$")

    df2 = ge.dataset.PandasDataset(
        {"id": [1, 2, 3, 4], "salary": [57000, 52000, 59000, 65000]}
    )

    df2.expect_column_values_to_match_regex("salary", "^[0-9]{4,6}$")

    df = df1.merge(df2, on="id")

    exp_m = [
        # No autoinspection as of 20180920
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'id'}},
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'name'}},
        # {'expectation_type': 'expect_column_to_exist',
        #  'kwargs': {'column': 'salary'}}
    ]

    # The merged data frame will:
    #
    #   1. Be a ge.dataset.PandaDataSet
    #   2. Have no expectations (no autoinspection is now default)

    assert isinstance(df, ge.dataset.PandasDataset)
    assert df.get_expectation_suite().expectations == exp_m


def test_ge_pandas_sampling():
    df = ge.dataset.PandasDataset(
        {
            "A": [1, 2, 3, 4],
            "B": [5, 6, 7, 8],
            "C": ["a", "b", "c", "d"],
            "D": ["e", "f", "g", "h"],
        }
    )

    # Put some simple expectations on the data frame
    df.profile(profiler=ColumnsExistProfiler)
    df.expect_column_values_to_be_in_set("A", [1, 2, 3, 4])
    df.expect_column_values_to_be_in_set("B", [5, 6, 7, 8])
    df.expect_column_values_to_be_in_set("C", ["a", "b", "c", "d"])
    df.expect_column_values_to_be_in_set("D", ["e", "f", "g", "h"])

    exp1 = df.get_expectation_suite().expectations

    # The sampled data frame should:
    #
    #   1. Be a ge.dataset.PandaDataSet
    #   2. Inherit ALL the expectations of the parent data frame

    samp1 = df.sample(n=2)
    assert isinstance(samp1, ge.dataset.PandasDataset)
    assert samp1.get_expectation_suite().expectations == exp1

    samp1 = df.sample(frac=0.25, replace=True)
    assert isinstance(samp1, ge.dataset.PandasDataset)
    assert samp1.get_expectation_suite().expectations == exp1

    # Change expectation on column "D", sample, and check expectations.
    # The failing expectation on column "D" is NOT automatically dropped
    # in the sample.
    df.expect_column_values_to_be_in_set("D", ["e", "f", "g", "x"])
    samp1 = df.sample(n=2)
    exp1 = expectationSuiteSchema.load(
        {
            "expectation_suite_name": "test",
            "expectations": [
                {
                    "expectation_type": "expect_column_to_exist",
                    "kwargs": {"column": "A"},
                },
                {
                    "expectation_type": "expect_column_to_exist",
                    "kwargs": {"column": "B"},
                },
                {
                    "expectation_type": "expect_column_to_exist",
                    "kwargs": {"column": "C"},
                },
                {
                    "expectation_type": "expect_column_to_exist",
                    "kwargs": {"column": "D"},
                },
                {
                    "expectation_type": "expect_column_values_to_be_in_set",
                    "kwargs": {"column": "A", "value_set": [1, 2, 3, 4]},
                },
                {
                    "expectation_type": "expect_column_values_to_be_in_set",
                    "kwargs": {"column": "B", "value_set": [5, 6, 7, 8]},
                },
                {
                    "expectation_type": "expect_column_values_to_be_in_set",
                    "kwargs": {"column": "C", "value_set": ["a", "b", "c", "d"]},
                },
                {
                    "expectation_type": "expect_column_values_to_be_in_set",
                    "kwargs": {"column": "D", "value_set": ["e", "f", "g", "x"]},
                },
            ],
        }
    )
    assert (
        samp1.get_expectation_suite(discard_failed_expectations=False).expectations
        == exp1.expectations
    )


def test_ge_pandas_subsetting():
    df = ge.dataset.PandasDataset(
        {
            "A": [1, 2, 3, 4],
            "B": [5, 6, 7, 8],
            "C": ["a", "b", "c", "d"],
            "D": ["e", "f", "g", "h"],
        }
    )

    # Put some simple expectations on the data frame
    df.expect_column_values_to_be_in_set("A", [1, 2, 3, 4])
    df.expect_column_values_to_be_in_set("B", [5, 6, 7, 8])
    df.expect_column_values_to_be_in_set("C", ["a", "b", "c", "d"])
    df.expect_column_values_to_be_in_set("D", ["e", "f", "g", "h"])

    # The subsetted data frame should:
    #
    #   1. Be a ge.dataset.PandaDataSet
    #   2. Inherit ALL the expectations of the parent data frame

    exp1 = df.get_expectation_suite().expectations

    sub1 = df[["A", "D"]]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df[["A"]]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df[:3]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df[1:2]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df[:-1]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df[-1:]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df.iloc[:3, 1:4]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1

    sub1 = df.loc[0:, "A":"B"]
    assert isinstance(sub1, ge.dataset.PandasDataset)
    assert sub1.get_expectation_suite().expectations == exp1


def test_ge_pandas_automatic_failure_removal():
    df = ge.dataset.PandasDataset(
        {
            "A": [1, 2, 3, 4],
            "B": [5, 6, 7, 8],
            "C": ["a", "b", "c", "d"],
            "D": ["e", "f", "g", "h"],
        }
    )

    # Put some simple expectations on the data frame
    df.profile(ge.profile.ColumnsExistProfiler)
    df.expect_column_values_to_be_in_set("A", [1, 2, 3, 4])
    df.expect_column_values_to_be_in_set("B", [5, 6, 7, 8])
    df.expect_column_values_to_be_in_set("C", ["w", "x", "y", "z"])
    df.expect_column_values_to_be_in_set("D", ["e", "f", "g", "h"])

    # First check that failing expectations are NOT automatically
    # dropped when sampling.
    # For this data frame, the expectation on column "C" above fails.
    exp1 = [
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "A"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "B"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "C"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "D"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "A", "value_set": [1, 2, 3, 4]},
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "B", "value_set": [5, 6, 7, 8]},
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "C", "value_set": ["w", "x", "y", "z"]},
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "D", "value_set": ["e", "f", "g", "h"]},
        ),
    ]
    samp1 = df.sample(n=2)
    assert (
        samp1.get_expectation_suite(discard_failed_expectations=False).expectations
        == exp1
    )

    # Now check subsetting to verify that failing expectations are NOT
    # automatically dropped when subsetting.
    sub1 = df[["A", "D"]]
    assert (
        samp1.get_expectation_suite(discard_failed_expectations=False).expectations
        == exp1
    )

    # Set property/attribute so that failing expectations are
    # automatically removed when sampling or subsetting.
    df.discard_subset_failing_expectations = True

    ###
    # Note: Order matters in this test, and a validationoperator may change order
    ###

    exp_samp = [
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "A"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "B"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "C"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "D"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "A", "value_set": [1, 2, 3, 4]},
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "B", "value_set": [5, 6, 7, 8]},
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "D", "value_set": ["e", "f", "g", "h"]},
        ),
    ]

    samp2 = df.sample(n=2)
    assert (
        samp2.get_expectation_suite(discard_failed_expectations=False).expectations
        == exp_samp
    )

    # Now check subsetting. In additional to the failure on column "C",
    # the expectations on column "B" now fail since column "B" doesn't
    # exist in the subset.
    sub2 = df[["A", "D"]]
    exp_sub = [
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "A"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "A", "value_set": [1, 2, 3, 4]},
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_to_exist", kwargs={"column": "D"}
        ),
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "D", "value_set": ["e", "f", "g", "h"]},
        ),
    ]
    assert (
        samp2.get_expectation_suite(discard_failed_expectations=False).expectations
        == exp_samp
    )


def test_subclass_pandas_subset_retains_subclass():
    """A subclass of PandasDataset should still be that subclass after a Pandas subsetting operation"""

    class CustomPandasDataset(ge.dataset.PandasDataset):
        @ge.dataset.MetaPandasDataset.column_map_expectation
        def expect_column_values_to_be_odd(self, column):
            return column.map(lambda x: x % 2)

        @ge.dataset.MetaPandasDataset.column_map_expectation
        def expectation_that_crashes_on_sixes(self, column):
            return column.map(lambda x: (x - 6) / 0 != "duck")

    df = CustomPandasDataset(
        {
            "all_odd": [1, 3, 5, 5, 5, 7, 9, 9, 9, 11],
            "mostly_odd": [1, 3, 5, 7, 9, 2, 4, 1, 3, 5],
            "all_even": [2, 4, 4, 6, 6, 6, 8, 8, 8, 8],
            "odd_missing": [1, 3, 5, None, None, None, None, 1, 3, None],
            "mixed_missing": [1, 3, 5, None, None, 2, 4, 1, 3, None],
            "all_missing": [None, None, None, None, None, None, None, None, None, None],
        }
    )

    df2 = df.sample(frac=0.5)
    assert type(df2) == type(df)

    def test_validate_map_expectation_on_categorical_column(self):
        """Map expectations should work on categorical columns"""

        D = ge.dataset.PandasDataset(
            {
                "cat_column_1": [
                    "cat_one",
                    "cat_two",
                    "cat_one",
                    "cat_two",
                    "cat_one",
                    "cat_two",
                    "cat_one",
                    "cat_two",
                ],
            }
        )

        D["cat_column_1"] = D["cat_column_1"].astype("category")

        D.set_default_expectation_argument("result_format", "COMPLETE")

        out = D.expect_column_value_lengths_to_equal("cat_column_1", 7)

        self.assertEqual(out["success"], True)


def test_pandas_deepcopy():
    import copy

    df = ge.dataset.PandasDataset({"a": [1, 2, 3]})
    df2 = copy.deepcopy(df)

    df["a"] = [2, 3, 4]

    # Our copied dataframe should not be affected
    assert df2.expect_column_to_exist("a").success == True
    assert list(df["a"]) == [2, 3, 4]
    assert list(df2["a"]) == [1, 2, 3]


def test_ge_value_count_of_object_dtype_column_with_mixed_types():
    """
    Having mixed type values in a object dtype column (e.g., strings and floats)
    used to raise a TypeError when sorting value_counts. This test verifies
    that the issue is fixed.
    """
    df = ge.dataset.PandasDataset(
        {"A": [1.5, 0.009, 0.5, "I am a string in an otherwise float column"],}
    )

    value_counts = df.get_column_value_counts("A")
    assert value_counts["I am a string in an otherwise float column"] == 1


def test_expect_values_to_be_of_type_list():
    """
    Having lists in a Pandas column used to raise a ValueError when parsing to
    see if any rows had missing values. This test verifies that the issue is fixed.
    """
    df = ge.dataset.PandasDataset({"A": [[1, 2], None, [4, 5], 6],})

    validation = df.expect_column_values_to_be_of_type("A", "list")
    assert not validation.success


def test_expect_values_quantiles_to_be_between():
    """
    Test that quantile bounds set to zero actually get interpreted as such. Zero
    used to be interpreted as None (and thus +-inf) and we'd get false negatives.
    """
    T = [
        ([1, 2, 3, 4, 5], [0.5], [[0, 0]], False),
        ([0, 0, 0, 0, 0], [0.5], [[0, 0]], True),
    ]

    for data, quantiles, value_ranges, success in T:
        df = ge.dataset.PandasDataset({"A": data})

        validation = df.expect_column_quantile_values_to_be_between(
            "A", {"quantiles": quantiles, "value_ranges": value_ranges,}
        )
        assert validation.success is success
